#!/usr/bin/perl 
# p8ldr version 0.9 -- date: Sept 15 2008
#
my $version = 0.9;
my $DEBUG = 0;
# Link object program and punch absolute binary papertape in 8+8 format
# or a relocatable load module in disk load format
#
# This loader can only be used to link object output of the p8asm assembler. 
# Loader output is either 
# - an absolute binary papertape image in 8+8 format, which can be loaded into 
#   the P800 simulator with the 'load' command or by starting the papertape 
#   bootstrap of the P800 simulator, or 
# - a relocatable load module in disk load format, which also can be loaded 
#   with the simulator 'load' command, or when available with an applicable 
#   P800 command. When a relocatable load module is produced, it is not allowed
#   to link any absolute code.
#
# todo:
# 1) link object modules provided as argument 2,3,..n
#    currently only the main object module is provided as argument 1;
#    the linker asks for a next module to link as long as externals are left
#    to be resolved.
#    >>>>>>(solved in version 0.7).
# 2) currently only an absolute load module is generated. an extension will be
#    to generate a relocatable load module as an alternative. 
#    >>>>>>(solved in version 0.5).
# 3) no multi segment object files are supported (object libraries) yet.
#    with this extension it will be possible that the loader selects object 
#    modules for linking, with entry points for which externals have to be 
#    resolved.
#    >>>>>>(solved in version 0.7).
#
# questions/assumptions:
# 1) is the length-word in the header of the relocatable disk load format
#    (word 3) counted in words or bytes? (this program assumes a length-count
#    in words
#    Found in a real load module is that this assumption is wrong:
#    word 3 required number of words is counted in bytes, . >>>>>> Adapted in version 0.8
# 2) format of the 12 word relocation table in a relocatable disk load format
#    sector? The following two assumptions are made in this program:
#    - a 1 indicates that a word must be relocated, a 0 indicates no relocation;
#    - the most significant bit of the 1st word in the relocation table (which
#      is word 188 in the sector, when start counting with word 0) relates to
#      the 1st code/data word in the sector (which then is word 0 in the sector)
#   Found in a real load module is that this assumption is ok! No adaption
#   required.
#
# - version 0.6   Adapted to process P800-C assembled files with symbols
#                 of max 10 chars long
#                 - allow leading underscore in entry points/external symbols
#                 - max symbol length is 10 chars (incl underscore)
#
# - version 0.7   Multiple modules can be specified on the command line.
#                 Library support: Libraries (extension .lib) are scanned 
#                 and only modules with entries for which externals are 
#                 to be resolved, are loaded from a library.
#
# - version 0.8   The relocational load module format has the 3rd word
#                 (required program length in words) counted in bytes.
#                 Although the load module is in the disk format as used to
#                 store load modules on disk, the 2 sector header words and the
#                 3 sector trailer words have a different content then on disk.
#                 The 2 sector header words each contain 0, and the 3 sector
#                 trailer words each contain /FFFF.
#
# - version 0.9   The absolute load module in 8+8 papertape format can now
#                 have a begin address of 0 
#
# Author: Theo Engel
# Info:   Info@theoengel.nl
#
# Usage:  p8ldr -[a|r] mainobjectprogram.obj [mod.obj|library.lib]..
#         The default is to produce an absolute load module in 8+8 format
#
#         Additional modules will be requested interactively, as long as
#         externals are to be resolved after loading the modules specified
#         on the command line
#
# Object code clusters (Assembler output)
# =======================================
# The assembler produces object code in the form of code clusters, which are
# input for the linker to generate the loadmodule. The cluster format (which is 
# NOT compatible with the cluster format of the original P800 assembler) is as
# follows:
#
# type 0: IDENT
#   1  1     10     1
# <som>0<NAME.....><eom>		         13 bytes
# - <NAME......> is the module name, 10 characters long
#
# type 1: ENTRY point definition
#   1  1     10       2     1     1
# <som>1<NAME......><ADDR><MODE><eom>    16 bytes
# - <NAME......> is the entry point name, 10 characters long
# - <ADDR>       is the binary address of the entry poiny, 2 bytes long
# - <MODE>       is mode of entry point address: A=absolute, R=relocatable
#
# type 3: CODE/DATA
#   1  1  2    1    2    1    1
# <som>3<ADDR><MA><WORD><MW><eom>        9 bytes
# - <ADDR>   is binary address of code/data word, 2 bytes long
# - <MA>     is mode of address: A=absolute, R=relocatable
# - <WORD>   is binary code/data word to be loaded at <ADDR>, 2 bytes long
# - <MW>     is is mode of the code/data word, 2 bytes long
#                                A=absolute address, 
#                                R=relocatable address
#                                0=code word
#
# type 4: EXTRN modification
#   1  1      6       2     1     1
# <som>4<NAME......><ADDR><MODE><eom>    16 bytes
# - <NAME......> is the name of the external, 10 characters long
# - <ADDR>       is the binary address of word where the external is used
#                in an operand expression, 2 bytes long
# - <MODE>       is mode of the word at <ADDR>: A=absolute, R=relocatable
#                When the symbol value of the external is known, that value is
#                added to the content of the word at <ADDR>
#
# type 7: END
#   1  1  2     1     2    1     1
# <som>7<ADDR><MODE><PRL><NERR><eom>     9 bytes
# - <ADDR>   is binary start address, 2 bytes long
# - <MODE>   is the mode of the start address,  A=absolute, R=relocatable
#                                               0=no start address defined
# - <PRL>    is module length in words, 2 bytes long
# - <NERR>   is the number of assembly errors, 1 byte long
#            (should be 0 when the module is linked)
#
# Format of the produced loadmodule
# =================================
# >> Absolute 8+8 papertape format (should be compatible with the origional one):
#
#	leader (0 actads), optional
#   rubout (octad 0xFF)
#   word count (16 bit, punched in 2 octads lbyte, rbyte)
#   begin address - 2 (16 bit hex word address punched in 2 octads lbyte, rbyte)
#   word count * data word (16 bit, punched in 2 octads lbyte, rbyte)
#   ...
#   ...
#   checksum (16 bit, punched in 2 octads lbyte, rbyte)
#   x-off   (octad 0x94)
#   trailer (0 actads), optional
#
#   wordcount is calculated as:
#       	(endaddress / 2) - (beginaddress / 2) + 1
#
#   1) both begin and end address must be wordaddresses (ls bit = 0)
#   2) the data part of the tape starts with content of begin address 
#      and ends with content of endaddress. 
#   3) the checksum is an exclusive or over the data words
#
# >> Relocatable disk load module format 
# (should be compatible with the origional one with respect to the 200 words
# in the record; different is that the records are written to "papertape" iso
# disk):
#
# Divided in records (sectors) of 200 words.
# The first 4 words of the 1st record contain:
# - start address of the load module (relocatable)
# - number of records/sectors of the load module
# - required memory length to load the load module (? in words ?)
# - an optional address of the entry point table (for debugging)
#
# The last 12 words contain relocation bits, while the other 184 words, starting
# at relative address 8, contain code words (if the program has that many words).
# The subsequent records/sectors (if there) each contain 188 code words and 12
# words with relocation bits.

use strict;
use bytes;

# hex digit conversion
my %cvdig;              
$cvdig{'0'}=0x0;
$cvdig{'1'}=0x1;
$cvdig{'2'}=0x2;
$cvdig{'3'}=0x3;
$cvdig{'4'}=0x4;
$cvdig{'5'}=0x5;
$cvdig{'6'}=0x6;
$cvdig{'7'}=0x7;
$cvdig{'8'}=0x8;
$cvdig{'9'}=0x9;
$cvdig{'A'}=0xa;
$cvdig{'B'}=0xb;
$cvdig{'C'}=0xc;
$cvdig{'D'}=0xd;
$cvdig{'E'}=0xe;
$cvdig{'F'}=0xf;
$cvdig{'a'}=0xa;
$cvdig{'b'}=0xb;
$cvdig{'c'}=0xc;
$cvdig{'d'}=0xd;
$cvdig{'e'}=0xe;
$cvdig{'f'}=0xf;

# ascii conversion table
my %ch;
$ch{0x20}=' ';
$ch{0x21}='!';
$ch{0x22}="\"";
$ch{0x23}="\#";
$ch{0x24}='$';
$ch{0x25}='%';
$ch{0x26}='&';
$ch{0x27}="\'";
$ch{0x28}='(';
$ch{0x29}=')';
$ch{0x2a}='*';
$ch{0x2b}='+';
$ch{0x2c}=',';
$ch{0x2d}='-';
$ch{0x2e}='.';
$ch{0x2f}="\/";
$ch{0x30}='0';
$ch{0x31}='1';
$ch{0x32}='2';
$ch{0x33}='3';
$ch{0x34}='4';
$ch{0x35}='5';
$ch{0x36}='6';
$ch{0x37}='7';
$ch{0x38}='8';
$ch{0x39}='9';
$ch{0x3a}="\:";
$ch{0x3b}="\;";
$ch{0x3c}='<';
$ch{0x3d}='=';
$ch{0x3e}='>';
$ch{0x3f}='?';
$ch{0x40}='@';
$ch{0x41}='A';
$ch{0x42}='B';
$ch{0x43}='C';
$ch{0x44}='D';
$ch{0x45}='E';
$ch{0x46}='F';
$ch{0x47}='G';
$ch{0x48}='H';
$ch{0x49}='I';
$ch{0x4a}='J';
$ch{0x4b}='K';
$ch{0x4c}='L';
$ch{0x4d}='M';
$ch{0x4e}='N';
$ch{0x4f}='O';
$ch{0x50}='P';
$ch{0x51}='Q';
$ch{0x52}='R';
$ch{0x53}='S';
$ch{0x54}='T';
$ch{0x55}='U';
$ch{0x56}='V';
$ch{0x57}='W';
$ch{0x58}='X';
$ch{0x59}='Y';
$ch{0x5a}='Z';
$ch{0x5b}='[';
$ch{0x5c}="\\";
$ch{0x5d}=']';
$ch{0x5e}='^';
$ch{0x5f}='_';

use constant MAXNAMELENGTH => 10;  # max symbol name length

my $linkmode = 'A';     # default linkmode
my $libmode  = 0;       # no library scan requested
my $fi="";              # input:  main module object file name
my $fo="";              # output: load module file name
my $narg=@ARGV;         # number of arguments on the command line
my $arg=0;              # argument to process

if($ARGV[0] eq '-a')
{
	$linkmode = 'A';
	$fi = $ARGV[1];
	$arg=2;
}
elsif($ARGV[0] eq '-r')
{
    $linkmode = 'R';
	$fi = $ARGV[1];
	$arg=2;
}
else
{
	$fi = $ARGV[0];
	$arg=1;
}

if ($fi eq "" || ( $fi !~ /\.obj/ ) )
{
    print "No main module filename, or wrong file extention (must be .obj)\n";
    exit 1;
}

my($fn,$ty) = split(/\./, $fi);
if($linkmode eq 'A')
{
	$fo = $fn . ".abs"; # punched absolute binary in 8+8 papertape load format
}
else
{
	$fo = $fn . ".rel"; # relocatable binary in disk load format
}

# nextchar/nextword
my $c =	0;	            # current char (8bit)
my $pc;		            # current char (7bit)
my $prevc;	            # previous char (8bit)
my $i;		            # character index in object file
my $w;                  # word

# tape format
my $null 	= 0;
my $som     = 0201;	    # start of codeblock
my $eom     = 0203;     # end of code block
my $rub     = 0377;     # start of tape
my $xoff    = 0224;     # end of tape
my $mskp 	= 0177;     # 7 bit mask
my $leader 	= "false";
my $end 	= "false";  # true if end block detected
my $blty	= 0;		# blocktype

#################### LOADER DATA #######################
my $begin=0;          # BEGIN ADDRESS, where loading is started
my $low=0xFFFF;       # LOW ADDRESS (lowest address of loaded code)
my $high=0;           # END ADDRESS (1st free address after loaded code)
my $start=0;          # START ADDRESS (last provided start address, if any)
my $loadaddr;         # load address of current module
my $mname;            # module name (in IDENT)
my $len;              # module length in words
my $prl=0;            # accumulated linked program length in words
my %sym;              # Symbol table: Entry Points/Externals (name address)
my %symref;           # Symbol references (addresses where symbol is in operand)
my $symbol;
my $mode;			  # A, R, X (0 in case of CODE, ! in case of error)
my $addr;             # address (text)
my $binaddr;          # binary address
my $word;             # code word (text)
my $binword;          # code word binary
my $nerr=0;           # number of detected errors during load
my $merr=0;           # number of assembly errors in module as specified in end block
my @M;                # memory array
my @R;                # relocator array (only used for a relocatable load module)
my $more='true';
########################################################

print "Relocating loader, version $version, Load object papertape: $fi\n";
if($linkmode eq 'A')
{
	print "An absolute papertape load module is requested as output: $fo\n";
	print "Provide load address (hex), + ENTER (or just ENTER in case of an absolute input tape): ";
    $begin = 0x800000;		
}
else
{
	print "A relocatable load module is requested as output: $fo\n";
	$begin = $low = 8;          # 1st 4 words used for start, #records, length, debug
}

while($linkmode eq 'A')
{
	read(STDIN,$c,1);
	if($cvdig{$c} ne undef)
	{
		$begin = $begin << 4;
		$begin = $begin + $cvdig{$c};
	}
	else
	{
		$begin = $begin & 0xFFFFFE;           # always load at word boundary
		if($begin == 0x800000)
		{
			print "\nNo start address is provided. Absolute loadmodule?\n";
			print "\nConfirm with y + ENTER to continue, or another character to enter a new load address: ";
			read(STDIN,$c,1);
			if($c eq "y" || $c eq "Y")
			{
				read(STDIN,$c,1); # cr
				last;
			}
			$begin = 0x800000;
			print "\nProvide a new load address (hex), ending with a ENTER: ";	
		}		
		elsif($begin > 0xFFFF)
		{
			print "\nLoad address bigger than 16 bit.";
			print "\nProvide a new load address (hex), ending with a ENTER: ";
			$begin = 0x800000;
		}
		else
		{
			print "\nProgram will start loading at: " . prhex(16,$begin);
			print "\nConfirm with y + ENTER to continue, or another character to enter a new load address: ";
			read(STDIN,$c,1);
			if($c eq "y" || $c eq "Y")
			{
				read(STDIN,$c,1); # cr
				last;
			}
			$begin = 0x800000;
			print "\nProvide a new load address (hex), ending with a ENTER: ";
		}
	}
}

$loadaddr = $begin;		                  # load address of current module = relocator

while($more eq 'true')
{
$leader = "false";
$end 	= "false";                        # true if end block detected	
$merr   = 0; 

print "Start reading: $fi\n";
open (IN, $fi) || die "cannot open inputfile: $fi: $!";
binmode IN, ":raw";

while( 1 )
{
		nextchar();
		#skip leader in case there
		if ($leader eq "false")
		{
			next if $c eq $null;
			$leader = "true";
			if($DEBUG) {print "Leader skipped\n";}
		}
		
		if($libmode && ($c eq $som))
		{
			if($DEBUG) {print "Scan Library\n";}
			scanlib();
			last;
		}
		else
		{
			# just load the specified module
			if( $c eq $som )
			{
				if($DEBUG) {print "\<som\>";}
				scanblock();
			}
			elsif( $c eq $eom )
			{
				if($DEBUG) {print "\<eom\>";}
				if($DEBUG) {print "\n";}
				if($end eq "true")
				{
					last;
				}
			}
			else
			{
				if($DEBUG) {print "\<";}
				if($DEBUG) {prx($c);}
				if($DEBUG) {print "\:$pc\>\n";}
			}
		}
}

if(! $libmode)
{
	print "\nModule $mname is loaded\n";
	close (IN) || die "can't close: $fi: $!";

	if($nerr != 0)
	{
		print "  $nerr load errors\n";
	}
	if($merr != 0)
	{
		print "  $merr assembly errors in: $mname\n";
	}
	if($nerr != 0 || $merr != 0)
	{
		print "Linking stopped because of errors\n";
		exit 1;
	}
	$high = $high + 2;        # point to 1st free address
	$loadaddr = $high;        # relocator for next module
}

print "\n";
print "BEGIN  " . prhex(16, $begin) . "\n";
print "LOW    " . prhex(16, $low) . " (lowest used address, sofar)\n";
print "HIGH   " . prhex(16, $high) . " (1st free address after used space)\n";

if($start == 0)
{
	print "START  ****\n\n";
}
else
{
	print "START  " . prhex(16, $start) . "\n\n";
}

my $sa;
my $sr;

$more = 'false';

# check whether all externals are defined
foreach $symbol (sort keys %sym)
{
   $sa = $sym{$symbol};
   $sr = $symref{$symbol};
   print "$symbol $sa $sr\n";
   
   if($sa eq '****')
   {
	   $more = 'true';
   }
}

# in case not all externals are defined, load an additional file
# (either a file specified on the command line, or asked for via the console)
if($more eq 'true')
{
	if(($narg-$arg) == 0)
	{
		# switch to interactive mode for requesting file names to load
		print "MORE\n";
		print "File name (+ ENTER): ";
		$fi = <STDIN>;
	}
	else
	{
		# process next argument
		$fi = $ARGV[$arg];
		$arg = $arg + 1;
	}

	if($fi =~ /\.lib/)
	{
		$libmode = 1;	# new library is requested to scan/load
	}
	elsif($fi =~ /\.obj/)
	{
		$libmode = 0;   # object module
	}
	else
	{
		print "File $fi with wrong extension specified; linking stopped\n";
		exit 1;
	}
}

} # end while ($more eq 'true')

print "\n";

# All modules loaded
# Resolve externals
my @ref;
my $i;
my $j;
my $sa;
my $sr;
my $rw;

foreach $symbol (sort keys %sym)
{
   $sa = $sym{$symbol};              # external address
   if($symref{$symbol} ne undef) 
   {
	    $sr = $symref{$symbol};      # reference to external
		@ref = split(/\,/, $sr);     # split reference into an array
		$i = @ref;
		# print "$symbol, $i references\n";
		$j = 0;
		while($j < $i)
		{
			$rw = $ref[$j];
			# print "$sa into ref: $rw\n";
			 $M[cvhex($rw)] =  $M[cvhex($rw)] + cvhex($sa);
			$j = $j + 1;
		}
   }
}

print "LOADING COMPLETE\n";

print "Length of the loadmodule in words (based on sum prl) : $prl (" . prhex(16,$prl) . " hex)\n";
$i=($high-$low)>>1;  # loadmodule length
print "Length of the loadmodule in words (based on low/high): $i (" . prhex(16,$i) . " hex)\n";

# Set header in case of a relocatable load module.
# If the header locations $M[0] $M[2] $M[4] and $M[6] are filled by the loader
# then there is an error
# M[0] = start address
# M[2} = number of sectors
# M[4] = program length (# of words required, specified in bytes)
# m[6] = symbol table address for debug (not used at this moment, so 0)
if($linkmode eq 'R')
{
    if( ($M[0] eq undef) || ($M[2] eq undef) || ($M[4] eq undef) || ($M[6] eq undef)
	    || ($low < $begin) )
	{
		$low=$low-8;         # extend load module with the 4 first words
		
		if($start == 0)
		{
			$M[0] = $begin;  # default start address
		}
		else
		{
			$M[0] = $start;
		}
		$R[0] = '*';         # set relocator for start address
		
        $i=(($high-$low)+1) >> 1;  # loadmodule length (incl header)
        print "Length of the loadmodule in words including header   : $i (" . prhex(16,$i) . " hex)\n";		
		$M[4] = $i << 1;     # set loadmodule length in bytes

		# 188 code words per record/sector + 12 words with relocator bits
		
	    $M[2] = int($i/188);   # number of  record/sectors in relocatable load module
		$i = $i % 188;         # sector part ?
		if($i > 0)             # yes
		{
			$M[2] = $M[2] + 1; # so an additional sector for the remainig part
		}
		print "The loadmodule requires $M[2] sectors on disk\n";  
	}
	else
	{
		print "ERROR: LOW < BEGIN, in a Relocatable LM, no code/data is alowed before BEGIN\n";
		exit 1;
	}
}

if($linkmode eq 'A')
{
	print "\nDump absolute linked program in hex on printer\n\n";
}
else
{
	print "\nDump relocatable linked program in hex on printer\n\n";
}

$i = $low;

while($i < $high)
{
	my $k = 0;
	my $c;
	my $s = "";
	$j = 0;
	print prhex(16,$i) . "  ";
	while($j < 16)        # 16 for 8 words, or 32 for 16 words
	{
		while($k < 4)
		{
			print prhex(16,$M[$i + $j]) . " ";  # word content
			if($linkmode eq 'R')                # relocator
			{
				if($R[$i+$j] eq '*')
				{
					print "< ";
				}
				else
				{
					print "  ";
				}
			}
			$c = ($M[$i + $j] >> 8) & 0xFF;     # character representation of content
			if($ch{$c} eq undef)
			{
				$c = ".";
			}
			else
			{
				$c = $ch{$c};
			}
			$s = $s . $c;
			$c = $M[$i + $j] & 0xFF;
			if($ch{$c} eq undef)
			{
				$c = ".";
			}
			else
			{
				$c = $ch{$c};
			}
			$s = $s . $c;
			$j = $j + 2;
			$k =$k + 1;
		}
		print " ";
		$k = 0;
	}
	print "$s\n";
	$i = $i + $j;
}

# output the linked load module
open (BIN,">:raw", $fo) || die "cannot open outputfile: $fo: $!";

if($linkmode eq 'A')
{
	print "\nPunch linked program in absolute 8+8 papertape load format: $fo \n\n";

	# 8+8 papertape format:
	#
	#	leader (0 actads), optional
	#   rubout (octad 0xFF)                                                         )
	#   word count (16 bit, punched in 2 octads lbyte, rbyte)                       )5 byte header
	#   begin address - 2 (16 bit hex word address punched in 2 octads lbyte, rbyte))
	#   ....
	#   word count * data word (16 bit, punched in 2 octads lbyte, rbyte)
	#   ...
	#   ...
	#   checksum (16 bit, punched in 2 octads lbyte, rbyte)                         )
	#   x-off   (octad 0x94)                                                        )3 byte trailer
	#   trailer (0 actads), optional
	#
	#   wordcount is calculated as:
	#       	(endaddress / 2) - (beginaddress / 2) + 1
	#
	#   1) both begin and end address must be wordaddresses (ls bit = 0)
	#   2) the data part of the tape starts with content of begin address 
	#      and ends with content of endaddress. 
	#   3) the checksum is an exclusive or over the data words
	#
	my $byte;
	my $w;
	my $chksum=0;

	punch($rub);                         # rubout
	$w = ($high >> 1) - ($low >> 1);     # word count (high = endaddr + 2)
	$byte = $w >> 8;
	punch($byte);
	$byte = $w & 0xFF;
	punch($byte);

	$w = $low-2;	                     # begin address - 2
	$byte = $w >> 8;
	punch($byte);
	$byte = $w & 0xFF;
	punch($byte);

	$i = $low;                           # begin address 

	while($i < $high)                    # content
	{
		$w = $M[$i];
		$chksum = $chksum ^ $w;
		$byte = $w >> 8;
		punch($byte);
		$byte = $w & 0xFF;
		punch($byte);
		$i = $i + 2;
	}

	$w=$chksum;                          # checksum
	$byte = $w >> 8;
	punch($byte);
	$byte = $w & 0xFF;
	punch($byte);
	punch($xoff);                        # xoff
}
else
{
	print "\nPunch linked program in relocatable disk load format: $fo \n\n";
	# on a P800 disk, a sector is 205 words long
	# 2 header  words
	# 200 data  words
	# 3 trailer words
	# this output can be assumed to be a P800 disk sector punched on papertape.
	# the header and trailer words are punched as words with content 0 cq /FFFF
	# the 200 data words contain the load module info:
	# max 188 code/data words + 12 words with relocation info
	# each code/data word is punched as <leftbyte><rightbyte>
	# the relocation bits start with the bit for code/data word 0 in the
	# most signification bit of the 1st relocation word.
	#
	my $i=$low;
	my $j=0;
	my $k=0;
	my $w;                     # word
	my $byte;                  # byte
	my $rb=0;                  # word with 16 relocator bits
	my @RB=();                 # words with relocator bits
	my $l=0;                   # max 12 relocator words / record/sector
	#
	
	while($i<$high)
	{
		# start a next record/sector
		$l=0;
		# output the 2 header words
		punch(0);punch(0); # w0
	    punch(0);punch(0); # w1
		$j=0;
		while($j<188)           # max 188 code/data words per record/sector
		{
            # start a next relocator word
			$rb=0;
			$k=0;
			while( ($k<16) && ($j < 188) && ($i < $high))       # max 16 relocators in a word
			{
				$w = $M[$i];     		# get content of next word
				$byte = $w >> 8;
				punch($byte);
				$byte = $w & 0xFF;
				punch($byte);           # and punch
				# has this word a relocator ?
				$rb = $rb << 1;
				if($R[$i] eq '*')
				{
					$rb = $rb | 1;
				}
				$k = $k + 1;            # next relocator
				$j = $j + 1;            # next word in sector
				$i = $i + 2;	        # next address				
			}
			# complete relocator word to 16 bit if not 16 bit yet
			while($k<16)
			{
				$rb=$rb << 1;
				$k=$k+1;
			}
			$RB[$l] = $rb;              # save relocator word
			$l=$l+1;
			if($l == 12)
			{
				last;
			}
		}
		# complete sector upto 188 words if not 188 words yet
		while($j < 188)
		{
			punch(0);punch(0);
			$j=$j+1;
		}
		# and output the 12 relocator words
		$l=0;
		while($l<12)
		{
			$w = $RB[$l];
			$byte = $w >> 8;
			punch($byte);
			$byte = $w & 0xFF;
			punch($byte);
			$l=$l+1;
		}
		# and output the last 3 sector trailer words
		punch(0xFF);punch(0xFF); # t0 (should be 0)
	    punch(0xFF);punch(0xFF); # t1 (should be 0)	
		punch(0xFF);punch(0xFF); # t2 (should be 0)
	}
}

close (BIN)  || die "can't close $fo: $!";

exit 0;

#############################################################################
sub scanblock {
	my $i;
	
	# enters afer reading som
	$blty = nextchar();	
	if($DEBUG) {print prdec(1, $blty);}
	if($blty == 0)
	{
		# IDENT
		$mname = "";
		if($DEBUG) { print "IDENT ";}
		$i=0;
		while($i<MAXNAMELENGTH)
		{
			nextchar();
			$mname = $mname . $pc;
			if($DEBUG) {print "$pc";}
			$i=$i+1;
		}
	}
	elsif($blty==1)
	{
		# ENTRY
		$symbol = "";                    # clear 
		if($DEBUG) {print "ENTRY ";}
		$i=0;
		while($i<MAXNAMELENGTH)
		{
			nextchar();
			if($DEBUG) {print "$pc";}
			$symbol = $symbol . $pc;     # symbol in text format
			$i=$i+1;
		}
		if($DEBUG) {print " ";}
		$binaddr = nextword();           # binary address
		$addr = prhex(16,$binaddr);   	 # address in text format
        if($DEBUG) {print "$addr";}
		nextchar();
		if($DEBUG) {print " $pc";}
		$mode = $pc;                     # mode in text format
		if($mode eq 'A')
		{
			# entry point has an absolute address
			if($sym{$symbol} eq undef || $sym{$symbol} eq "****" )
			{
				$sym{$symbol} = $addr;
			}
			else
			{
				error("Double defined Entry point: $symbol");
			}
		}
		elsif($mode eq 'R')
		{
			# entry point has a relocatable address
			$binaddr = $binaddr + $loadaddr;         # relocate
			if($sym{$symbol} eq undef || $sym{$symbol} eq "****")
			{
				$sym{$symbol} = prhex(16,$binaddr); 
			}
			else
			{
				error("Double defined Entry point: $symbol");
			}			
		}
		else
		{
			error("Entry point $symbol with address in wrong mode $mode");
		}
	}
	elsif($blty==4)
	{
		# EXTRN
		$symbol = "";
		if($DEBUG) {print "EXTRN ";}
		$i=0;
		while($i<MAXNAMELENGTH)
		{
			nextchar();
			if($DEBUG) {print "$pc";}
			$symbol = $symbol . $pc;                # symbol in text format
			$i=$i+1;
		}
		if($DEBUG) {print " ";}
		$binaddr = nextword();          # binary address
		$addr = prhex(16,$binaddr);   	# address in text format
        if($DEBUG) {print "$addr";}
		nextchar();
		if($DEBUG) {print " $pc";}
		$mode = $pc;                    # mode in text format
		if($sym{$symbol} eq undef)
		{
			$sym{$symbol} = "****";
		}
		
		if($mode eq 'A')
		{
			# external reference address is absolute
			if($symref{$symbol} eq undef)
			{
				$symref{$symbol} = $addr; 
			}
			else
			{
				$symref{$symbol} = $symref{$symbol} . "," . $addr; 
			}
		}
		elsif($mode eq 'R')
		{
			# external reference address is relocatable
			$binaddr = $binaddr + $loadaddr;         # relocate
			if($symref{$symbol} eq undef)
			{
				$symref{$symbol} =  prhex(16,$binaddr); 
			}
			else
			{
				$symref{$symbol} = $symref{$symbol} . "," . prhex(16,$binaddr); 
			}
		}
		else
		{
			error("External $symbol with refence address in wrong mode $mode");
		}
	}
	elsif($blty==3)
	{
		# CODE/DATA
		if($DEBUG) {print "CODE  ";}
		$binaddr = nextword();
		$addr = prhex(16,$binaddr);
		if($DEBUG) {print $addr;}	# address
		nextchar();
		if($DEBUG) {print " $pc ";}
		$mode = $pc;
		if($mode eq "A")
		{
			if($begin == 0)
			{
				$begin = $binaddr;
			}
			if($linkmode eq 'R')
			{
				error("Absolute code at address $addr not allowed in a relocatable module");
			}
		}
		elsif($mode eq "R")
		{
			$binaddr = $binaddr + $loadaddr; # relocate
		}
		else
		{
			error("Address $addr has wrong mode $mode");
		}
		$binword = nextword();
		$word = prhex(16,$binword);
		if($DEBUG) {print $word;}	# content
		nextchar();
		if($DEBUG) {print " $pc ";}
		$mode = $pc;
		if($mode eq "A" || $mode eq "0")
		{
			;
		}
		elsif($mode eq "R")
		{
			$binword = $binword + $loadaddr; # relocate
			if($linkmode eq 'R')
			{
				$R[$binaddr] = '*';          # set relocator
			}
		}
		else
		{
			error("Content at $addr has wrong mode $mode");
		}
		# store in 'memory'
		$M[$binaddr] = $binword;
		# update LOW and HIGH
		if($binaddr < $low)
		{
			$low = $binaddr;
		}
		if($binaddr > $high)
		{
			$high = $binaddr;
		}		
	}
	elsif($blty==7)
	{
		# END
		if($DEBUG) {print "END   ";}
		$binaddr = nextword();
		$addr = prhex(16,$binaddr);
		if($DEBUG) {print $addr;}
		nextchar();            
		if($DEBUG) {print " $pc ";}          # mode
		$mode = $pc;
		if($pc ne "0")
		{
			# start address is specified
			if($mode eq "A")
			{
				$start = $binaddr;
				if($linkmode eq 'R')
				{
					error("Absolute start address $addr not allowed in a relocatable module");
				}
			}
			elsif($mode eq "R")
			{
				$start = $binaddr + $loadaddr;
			}
			else
			{
				error("Start address in END-block has wrong mode $mode");
			}
		}
		$len = nextword();
		if($DEBUG) {print prhex(16,$len);}	# module length in words
		$prl = $prl + $len;     # accumulated program length in words
		if($DEBUG) {print " ";}
		$merr = nextchar();
		if($DEBUG) {print prhex(8,$merr);}   # number of assembly errors
		$end = 'true';
	}
}
###############################################################################
# scanlib
# scans a library and only loads the modules of that library for which externals
# can be resolved. So in case a module is scanned in the library, which
# has one or more entry points which exist as external in the symbol table, that
# library module is loaded. The loaded library module itself can have externals 
# to another module. That module then be presented to the loader after the 1st 
# loaded library module. To prevent multiple loads of a library, modules must
# be entered in a library so, that a module with particalar entry points is 
# after modules in the library which has those entry points as externals.

my $modname=""; # name of the currently scanned module
my %mentries;   # entry points of the currently scanned module
my %mextrns;    # externals of the currently scanned module
my %mnames;     # names of modules in the library that must be loaded
my $eof=0;
my $lflag=0;

sub scanlib {
	my $blty; my $i;
	%mnames=();  

	# enters afer reading som	
	
	# Pass 1: scan library to collect the modules to load from the library
	while(! $eof)
	{
		$blty = nextchar();	   
		if($blty == 0)
		{
			# IDENT	
			$modname = "";
			$lflag   = 0;
			if($DEBUG) {print "IDENT ";}
			# collect name, entry points and externals of the current module
			$i=0;
			while($i<MAXNAMELENGTH)
			{
				nextchar();
				$modname = $modname . $pc;
				if($DEBUG) {print "$pc";}			
				$i=$i+1;
			}
	        %mentries=();
	        %mextrns=();			
			while(1)
			{
				nextchar();
				if($c eq $som)
				{
					$blty = nextchar();
					if($blty == 1)
					{
						# ENTRY
						$symbol = "";          # clear 	
						if($DEBUG) {print "ENTRY ";}
						$i=0;
						while($i<MAXNAMELENGTH)
						{
							nextchar();
							$symbol = $symbol . $pc;
							if($DEBUG) {print "$pc";}			
							$i=$i+1;
						}			
						$mentries{$symbol} = " ";	# enter entry point in hash
					}
					elsif($blty == 4)
					{
						# EXTRN
						$symbol = "";          # clear 	
						if($DEBUG) {print "EXTRN ";}
						$i=0;
						while($i<MAXNAMELENGTH)
						{
							nextchar();
							$symbol = $symbol . $pc;
							if($DEBUG) {print "$pc";}			
							$i=$i+1;
						}			
						$mextrns{$symbol} = " ";	# enter external in hash
					}
					elsif($blty == 7)
					{
						# END
						if($DEBUG) {print "END   ";}
						if($DEBUG) {print "\n";}
						# end of library module
						# module data is collected
						last;
					}
				}
			}
			# check whether this module must be loaded
			foreach $symbol (keys %mentries)
			{
				if($sym{$symbol} eq '****')
				{
					# yes libmodules entry point lives as external in the symbol table
					# store libmodule name in the hash with modules to load
					$mnames{$modname} = " ";
					$lflag = 1;
					last;
				}
			}
			if($lflag)
			{
				# put the libmodules externals in the symbol table (if not there yet)
				foreach $symbol (keys %mextrns)
				{
					if($sym{$symbol} eq undef)
					{
						$sym{$symbol} = "****";
					}
				}
			}
		}

		# skip to next block
		while(1)
		{
			if(nextchar() == -1)
			{
				$eof=1;
				last;
			}
			elsif($c eq $som )
			{
				last;
			}
		}
	}
	
	# end of pass 1
	close(IN);
	$i=%mnames;
	if($i==0)
	{
		print "No modules to load from library $fi\n";
		return;
	}
	print "The following modules must be loaded from the library $fi\n";
	foreach $symbol (keys %mnames)
	{
		print "Module: $symbol \n";
	}
	print "Start reading: $fi (pass 2)\n";
	open (IN, $fi) || die "cannot open inputfile: $fi: $!";
	binmode IN, ":raw";	
	
	while(1)
	{
		if(nextchar() == -1)
		{
			last; # eof
		}
		#skip leader in case there
		if ($leader eq "false")
		{
			next if $c eq $null;
			$leader = "true";
			if($DEBUG) {print "Leader skipped\n";}
		}
		
		if($c eq $som )
		{		
			$blty = nextchar();	   
			if($blty == 0)
			{
				# IDENT	
				$modname = "";
				if($DEBUG) {print "IDENT ";}
				$i=0;
				while($i<MAXNAMELENGTH)
				{	
					nextchar();
					$modname = $modname . $pc;
					if($DEBUG) {print "$pc";}		
					$i=$i+1;
				}
				if($mnames{$modname} ne undef)
				{
					# load this module!
					if($DEBUG) {print "LOAD!";}
					$end = "false";
					$merr   =0; 
					while(1)
					{
						nextchar();  # som
						if( $c eq $som )
						{
							if($DEBUG) {print "\<som\>";}
							scanblock();
						}
						elsif( $c eq $eom )
						{
							if($DEBUG) {print "\<eom\>";}
							if($DEBUG) {print "\n";}
							if($end eq "true")
							{
								last;
							}
						}
						else
						{
							if($DEBUG) {print "\<";}
							if($DEBUG) {prx($c);}
							if($DEBUG) {print "\:$pc\>\n";}
						}
					}
					print "Library Module $modname is loaded\n";
					if($nerr != 0)
					{
						print "  $nerr load errors\n";
					}
					if($merr != 0)
					{
						print "  $merr assembly errors in: $mname\n";
					}
					if($nerr != 0 || $merr != 0)
					{
						print "Linking stopped because of errors\n";
						exit 1;
					}
					$high = $high + 2;         # point to 1st free address
					$loadaddr = $high;         # relocator for next module
				}
				if($DEBUG) {print "\n";}
			}
		} # c eq <som>
	} # end pass 2
} # end scan lib

###############################################################################

sub nextchar {
	    if(read(IN,$c,1) == 0)
		{
			return -1;						#EOF
		}
		$prevc = $c; 				        # save previous char
		$c = unpack("C", $c); 	            # get new current char
		$pc= $c & $mskp;
		$pc= chr($pc);				        # current char in printable form
		$i=$i+1;
		return ($c & 0xFF);
}

sub nextword {
	$w = nextchar();
	$w = $w << 8;
	nextchar();
	$w = $w + $c;
	return $w;
}

sub prx { 
my $v = shift;
my $s = sprintf ("%x", $v);
print $s;
}

sub cvhex {
	# convert 4 digit hex string to binary
	my $s = shift;
	my $r = 0;
	my $c;
	$c=substr($s,0,1);
	$c = $cvdig{$c};
	$r = ($r << 4) + $c;
	$c=substr($s,1,1);
	$c = $cvdig{$c};
	$r = ($r << 4) + $c;
	$c=substr($s,2,1);
	$c = $cvdig{$c};
	$r = ($r << 4) + $c;
	$c=substr($s,3,1);
	$c = $cvdig{$c};
	$r = ($r << 4) + $c;
	return $r;
}

sub prhex {
	# print hex
	# $n is number of bits
	# $v is value
	my $n = shift;
	my $v = shift;
	my $s;
	if($n == 0)
	{
		return "";
	}
	elsif($n<5)
	{
		$s = sprintf("%01X", $v);	
	}
	elsif($n<9)
	{
		$s = sprintf("%02X", $v);
	}
	elsif($n<13)
	{		
		$s = sprintf("%03X", $v);
	}
	elsif($n<17)
	{		
		$s = sprintf("%04X", $v);
	}
	elsif($n<21)
	{
		$s = sprintf("%05X", $v);
	}
	elsif($n<25)
	{						
		$s = sprintf("%06X", $v);
	}
	elsif($n<29)
	{		
		$s = sprintf("%07X", $v);
	}
	elsif($n<33)
	{
		$s = sprintf("%08X", $v);
	}
	# print "$s ";
	return $s;
}

sub rtrim {
	my $s = shift;
	my $l = length($s);
	my $i = $l;
	if ($l > 0)
	{
		while($i > 0)
		{
			if((substr($s,$i-1,1)) ne " ")
			{ 
				last;
			}
			else
			{
				$i=$i-1;
			}
		}
	}
	else
	{
		return "";
	}
	if($i == 0)
	{
		return "";
	}
	else
	{
		return substr($s,0,$i);
	}
}

sub prdec
{
	# print number
	# $n is number of digits (string)
	# $v is value
	my $n = shift;
	my $v = shift;
	my $s;
	my $format="\%0" . $n . "d";
	$s = sprintf($format, $v);
	# print "$s ";
	return "  $s  ";
}

sub error
{
	my $es = shift;
	$nerr = $nerr + 1;
	print "****ERROR**** $es\n";
	
}

sub punch {
    my($a) = @_;
    my $byte;
    $byte = pack("C", ($a & 0xFF) );
    print BIN $byte;
}
